@@ -5,9 +5,10 @@
  * a Linking Exception. For full terms see the included COPYING file.
  */
 
-#include "common.h"
 #include "vector.h"
 
+#include "integer.h"
+
 /* In elements, not bytes */
 #define MIN_ALLOCSIZE	8
 
@@ -29,14 +30,9 @@ GIT_INLINE(size_t) compute_new_size(git_vector *v)
 
 GIT_INLINE(int) resize_vector(git_vector *v, size_t new_size)
 {
-	size_t new_bytes = new_size * sizeof(void *);
 	void *new_contents;
 
-	/* Check for overflow */
-	if (new_bytes / sizeof(void *) != new_size)
-		GITERR_CHECK_ALLOC(NULL);
-
-	new_contents = git__realloc(v->contents, new_bytes);
+	new_contents = git__reallocarray(v->contents, new_size, sizeof(void *));
 	GITERR_CHECK_ALLOC(new_contents);
 
 	v->_alloc_size = new_size;
@@ -45,13 +41,20 @@ GIT_INLINE(int) resize_vector(git_vector *v, size_t new_size)
 	return 0;
 }
 
+int git_vector_size_hint(git_vector *v, size_t size_hint)
+{
+	if (v->_alloc_size >= size_hint)
+		return 0;
+	return resize_vector(v, size_hint);
+}
+
 int git_vector_dup(git_vector *v, const git_vector *src, git_vector_cmp cmp)
 {
 	size_t bytes;
 
 	assert(v && src);
 
-	bytes = src->length * sizeof(void *);
+	GITERR_CHECK_ALLOC_MULTIPLY(&bytes, src->length, sizeof(void *));
 
 	v->_alloc_size = src->length;
 	v->_cmp = cmp ? cmp : src->_cmp;
@@ -328,6 +331,47 @@ int git_vector_resize_to(git_vector *v, size_t new_length)
 	return 0;
 }
 
+int git_vector_insert_null(git_vector *v, size_t idx, size_t insert_len)
+{
+	size_t new_length;
+
+	assert(insert_len > 0 && idx <= v->length);
+
+	GITERR_CHECK_ALLOC_ADD(&new_length, v->length, insert_len);
+
+	if (new_length > v->_alloc_size && resize_vector(v, new_length) < 0)
+		return -1;
+
+	memmove(&v->contents[idx + insert_len], &v->contents[idx],
+		sizeof(void *) * (v->length - idx));
+	memset(&v->contents[idx], 0, sizeof(void *) * insert_len);
+
+	v->length = new_length;
+	return 0;
+}
+
+int git_vector_remove_range(git_vector *v, size_t idx, size_t remove_len)
+{
+	size_t new_length = v->length - remove_len;
+	size_t end_idx = 0;
+	
+	assert(remove_len > 0);
+
+	if (git__add_sizet_overflow(&end_idx, idx, remove_len))
+		assert(0);
+
+	assert(end_idx <= v->length);
+
+	if (end_idx < v->length)
+		memmove(&v->contents[idx], &v->contents[end_idx],
+			sizeof(void *) * (v->length - end_idx));
+
+	memset(&v->contents[new_length], 0, sizeof(void *) * remove_len);
+
+	v->length = new_length;
+	return 0;
+}
+
 int git_vector_set(void **old, git_vector *v, size_t position, void *value)
 {
 	if (position + 1 > v->length) {
@@ -357,3 +401,22 @@ int git_vector_verify_sorted(const git_vector *v)
 
 	return 0;
 }
+
+void git_vector_reverse(git_vector *v)
+{
+	size_t a, b;
+
+	if (v->length == 0)
+		return;
+
+	a = 0;
+	b = v->length - 1;
+
+	while (a < b) {
+		void *tmp = v->contents[a];
+		v->contents[a] = v->contents[b];
+		v->contents[b] = tmp;
+		a++;
+		b--;
+	}
+}
